{ *****************************************************
                  ThgCustomChart Component

  ThgCustomChart is a custom base class for TXYChart and
  other XY, scatter, line and bar charts.

                    Paul Warren
           HomeGrown Software Development
         (c) 1997 Langley British Columbia.
                  (604) 856-6523
           e-mail:  hg_soft@uniserve.com
     Home page: http://users.uniserve.com/~hg_soft
  ***************************************************** }

unit Charts;

interface

uses
  SysUtils, WinTypes, WinProcs, Messages, Classes, Graphics, Controls,
  Forms, Dialogs, StdCtrls, Math;

type
  TTitleStr = string[35];
  TGridType = (gtNone, gtHorz, gtVert, gtBoth);

  ThgCustomChart = class(TGraphicControl)
  private
    { private declarations }
    FTitle: TTitleStr;
    FTitleFont: TFont;
    FXLabel: TTitleStr;
    FYLabel: TTitleStr;
    FXYLabelFont: TFont;
    FGrid: TGridType;
    FBackGround: TColor;
    FChartColor: TColor;
    FBorderWidth: integer;
    FMinX: Double;
    FMinY: Double;
    FMaxX: Double;
    FMaxY: Double;
    OffScreen: TBitMap;
    function GetChartRect: TRect; virtual;
    function GetDrawRect: TRect; virtual; abstract;
    procedure SetTitle(Value: TTitleStr);
    procedure SetTitleFont(Value: TFont);
    procedure SetXLabel(Value: TTitleStr);
    procedure SetYLabel(Value: TTitleStr);
    procedure SetXYLabelFont(Value: TFont);
    procedure SetGrid(Value: TGridType);
    procedure SetBackGround(Value: TColor);
    procedure SetChartColor(Value: TColor);
    procedure SetBorderWidth(Value: integer);
  protected
    { protected declarations }
    procedure Paint; override;
    procedure DrawGrid(Rect: TRect); virtual; abstract;
    procedure DrawXScale; virtual; abstract;
    procedure DrawYScale; virtual; abstract;
    procedure DrawDataPoints; virtual; abstract;
    procedure ScaleChart; virtual; abstract;
    property ChartRect: TRect read GetChartRect;
  public
    { public declarations }
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    property Title: TTitleStr read FTitle write SetTitle;
    property TitleFont: TFont read FTitleFont write SetTitleFont;
    property XLabel: TTitleStr read FXLabel write SetXLabel;
    property YLabel: TTitleStr read FYLabel write SetYLabel;
    property XYLabelFont: TFont read FXYLabelFont write SetXYLabelFont;
    property Grid: TGridType read FGrid write SetGrid default gtNone;
    property BackGround: TColor read FBackGround write SetBackGround default clSilver;
    property ChartColor: TColor read FChartColor write SetChartColor default clWhite;
    property BorderWidth: integer read FBorderWidth write SetBorderWidth default 1;
    property MinX: Double read FMinX write FMinX;
    property MaxX: Double read FMaxX write FMaxX;
    property MinY: Double read FMinY write FMinY;
    property MaxY: Double read FMaxY write FMaxY;
  published
    { published declarations }
  end;

  TDataPoint = class(TObject)
    X, Y: Double;
  end;

  TChartDataList = class(TStringList)
  private
    { private declarations }
  public
    { public declarations }
    destructor Destroy; override;
    function AddDataPoint(Point: TDataPoint): integer;
    function GetDataPoint(index: integer): TDataPoint;
  end;

  TChartType = (ctScatter, ctLine, ctBar, ctXY);
  TAxisType = (atXAxis, atYAxis);
  TNeedLines = procedure(Sender: TObject; var Value: Double;
    var AColor: TColor; var Finished: boolean) of object;
  TScaling = procedure(Sender: TObject; var MinX, MaxX, MinY, MaxY: Double) of object;
  TDrawScales = procedure(Sender: TObject; Axis: TAxisType; var Text: string) of object;

  TXYChart = class(ThgCustomChart)
  private
    { private declarations }
    FChartType: TChartType;
    FDataList: TChartDataList;
    FOnNeedLinesX: TNeedLines;
    FOnNeedLinesY: TNeedLines;
    FOnScaling: TScaling;
    FOnDrawScales: TDrawScales;
    function GetChartRect: TRect; override;
    function GetDrawRect: TRect; override;
    function GetBarWidth: integer;
    procedure SetChartType(Value: TChartType);
    procedure SetData(Value: TChartDataList);
  protected
    { protected declarations }
    procedure ScaleChart; override;
    function NormalizePtX(APoint: Double): Double;
    function NormalizePtY(APoint: Double): Double;
    procedure DrawGrid(Rect: TRect); override;
    procedure DrawXScale; override;
    procedure DrawYScale; override;
    procedure DrawDataPoints; override;
    property BarWidth: integer read GetBarWidth;
  public
    { public declarations }
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure AddData(X, Y: Double);
    procedure Plot;
    procedure ClearData;
    property Data: TChartDataList read FDataList write SetData;
  published
    { published declarations }
    property ChartType: TChartType read FChartType write SetChartType;
    property OnNeedLinesX: TNeedLines read FOnNeedLinesX write FOnNeedLinesX;
    property OnNeedLinesY: TNeedLines read FOnNeedLinesY write FOnNeedLinesY;
    property OnScaling: TScaling read FOnScaling write FOnScaling;
    property OnDrawScales: TDrawScales read FOnDrawScales write FOnDrawScales;
    property Align;
    property BackGround;
    property BorderWidth;
    property ChartColor;
    property Font;
    property Grid;
    property Title;
    property TitleFont;
    property XLabel;
    property YLabel;
    property XYLabelFont;
  end;

procedure Register;

implementation

{$IFDEF WIN32}
  {$R CHARTS.D32}
{$ELSE}
  {$R CHARTS.D16}
{$ENDIF}

{ ThgCustomChart }
constructor ThgCustomChart.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  ControlStyle := ControlStyle + [csOpaque];
  Width := 147;
  Height := 105;
  FTitle := 'HomeGrown Charts';
  FTitleFont := TFont.Create;
  FTitleFont.Name := 'arial';
  FTitleFont.Style := [fsbold];
  FTitleFont.Size := 11;
  FTitleFont.Color := clRed;
  FXLabel := 'X-Axis';
  FYLabel := 'Y-Axis';
  FXYLabelFont := TFont.Create;
  FXYLabelFont.Name := 'arial';
  FXYLabelFont.Size := 8;
  FGrid := gtNone;
  FBackGround := clSilver;
  FChartColor := clWhite;
  FBorderWidth := 1;
  FMinX := 0;
  FMaxX := 50;
  FMinY := 0;
  FMaxY := 50;
  OffScreen := TBitMap.Create;
end;

destructor ThgCustomChart.Destroy;
begin
   FTitleFont.Free;
   FXYLabelFont.Free;
   OffScreen.Free;
   inherited Destroy;
end;

{ property access routines }
procedure ThgCustomChart.SetTitle(Value: TTitleStr);
begin
  FTitle := Value;
  Refresh;
end;

procedure ThgCustomChart.SetTitleFont(Value: TFont);
begin
  FTitleFont.Assign(Value);
  Refresh;
end;

procedure ThgCustomChart.SetXLabel(Value: TTitleStr);
begin
  FXLabel := Value;
  Refresh;
end;

procedure ThgCustomChart.SetYLabel(Value: TTitleStr);
begin
  FYLabel := Value;
  Refresh;
end;

procedure ThgCustomChart.SetXYLabelFont(Value: TFont);
begin
  FXYLabelFont.Assign(Value);
  Refresh;
end;

procedure ThgCustomChart.SetGrid(Value: TGridType);
begin
  if Value <> FGrid then
  begin
    FGrid := Value;
    Refresh;
  end;
end;

procedure ThgCustomChart.SetBackGround(Value: TColor);
begin
  if Value <> FBackGround then
  begin
    FBackGround := Value;
    Refresh;
  end;
end;

procedure ThgCustomChart.SetChartColor(Value: TColor);
begin
  if Value <> FChartColor then
  begin
    FChartColor := Value;
    Refresh;
  end;
end;

procedure ThgCustomChart.SetBorderWidth(Value: integer);
begin
  if Value <> FBorderWidth then
  begin
    FBorderWidth := Value;
    if FBorderWidth < 1 then FBorderWidth := 1;
    if FBorderWidth > 4 then FBorderWidth := 4;
    Refresh;
  end;
end;

{ GetChartRect method }
function ThgCustomChart.GetChartRect: TRect;
begin
  with OffScreen do
  begin
    { calulate the TRect for the whole chart }
    Canvas.Font := TitleFont;
    Result.Top := Round(Canvas.TextHeight(FTitle) * 1.2);
    Canvas.Font := XYLabelFont;
    Result.Left := (Canvas.TextHeight(FTitle)*2);
    Result.Right := Width - Canvas.TextWidth('10000');
    Result.Bottom := Height - Round(Canvas.TextHeight(FTitle) * 2.4);
    Canvas.Font := Font;
  end;
end;

{ Paint method }
procedure ThgCustomChart.Paint;
var
  hFontNew: HFONT;
  hFontOld: HFONT;
  p: string;
begin
  { paint handles all the drawing unrelated to the actual plot
    ie. title, color, border etc. }
  OffScreen.Width := ClientWidth;
  OffScreen.Height := ClientHeight;
  Font := Font;
  with OffScreen.Canvas do
  begin
    { draw background }
    Brush.Style := bsSolid;
    Brush.Color := FBackGround;
    FillRect(ClientRect);

    { draw title }
    Font := TitleFont;
    TextOut((Width-TextWidth(FTitle)) div 2, 1, FTitle);

    { draw x-axis label }
    Font := XYLabelFont;
    TextOut((Width-TextWidth(FXLabel)) div 2, Height-TextHeight(FXLabel), FXLabel);

    { draw y-axis  label }
    p := Font.Name + #0; { PChar }
    { Creation of the object font and angle }
    hFontNew:= CreateFont(
      { Height      } Font.Height,
      { Width       } 0,
      { Escapement  } 900, { angle of slant of text }
      { Orientation } 0, { expressed in tenth of degree  }
      { Weight      } ord(fsBold in Font.Style) * FW_BOLD,
      { Italic      } ord(fsItalic     in Font.Style),
      { Underline   } ord(fsUnderLine  in Font.Style),
      { StrikeOut   } ord(fsStrikeOut  in Font.Style),
      { Charset     } DEFAULT_CHARSET,
      { OutputPrec. } OUT_DEFAULT_PRECIS,
      { ClipPrec.   } CLIP_DEFAULT_PRECIS,
      { Quality     } DEFAULT_QUALITY,
      { Pitch       } DEFAULT_PITCH + FF_DONTCARE,
      { FaceName    } @p[1]
    );
    { store old font and select new }
    hFontOld := SelectObject(Handle, hFontNew);
    { output text in new font }
    TextOut(1, (Height+TextHeight(FYLabel)) div 2, FYLabel);
    { return the old font }
    SelectObject(Handle, hFontOld);
    { delete the new font }
    DeleteObject(hFontNew);

    { scale the chart - abstract }
    ScaleChart;

    { draw border }
    Brush.Style := bsSolid;
    Brush.Color := FChartColor;
    Pen.Width := FBorderWidth;
    RoundRect(ChartRect.Left, ChartRect.Top, ChartRect.Right,
      ChartRect.Bottom, 15, 15);
    Pen.Width := 1;
    Font.Color := clBlack;

    { draw Grid }
    DrawGrid(ChartRect);

    { draw axis scales - abstract }
    DrawXScale;
    DrawYScale;

    { draw data points - abstract }
    DrawDataPoints;
  end;
  Canvas.Draw(0, 0, OffScreen);
end;

{ TChartDataList - data structure for the chart data }
destructor TChartDataList.Destroy;
var
  Temp: TObject;
begin
  while Count <> 0 do
  begin
    { make Temp := first Item }
    Temp := Objects[0];
    { free it }
    Temp.Free;
    { delete it }
    Delete(0);
  end;
  { call inherited }
  inherited Destroy;  
end;

function TChartDataList.AddDataPoint(Point: TDataPoint): integer;
begin
  Result := AddObject('', Point);
end;

function TChartDataList.GetDataPoint(index: integer): TDataPoint;
begin
  Result := Objects[index] as TDataPoint;
end;

{ TXYChart }
constructor TXYChart.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  { create the data list }
  FDataList := TChartDataList.Create;
end;

destructor TXYChart.Destroy;
begin
  { free the data list }
  FDataList.Free;
  inherited Destroy;
end;

procedure TXYChart.AddData(X, Y: Double);
var
  P: TDataPoint;
begin
  { create data object }
  P := TDataPoint.Create;
  P.X := X; P.Y := Y;
  { add it to the list }
  FDataList.AddDataPoint(P);
  { don't need to free - the list does that }
end;

procedure TXYChart.SetChartType(Value: TChartType);
begin
  if Value <> FChartType then
  begin
    FChartType := Value;
    Refresh;
  end;
end;

procedure TXYChart.ClearData;
begin
  if FDataList.Count <> 0 then FDataList.Clear;
end;

procedure TXYChart.Plot;
begin
  Refresh;
end;

procedure TXYChart.SetData(Value: TChartDataList);
begin
  FDataList.Free;
  FDataList.Assign(Value);
  Refresh;
end;

{ GetChartRect method }
function TXYChart.GetChartRect: TRect;
begin
  Result := inherited GetChartRect;
  Result.Left := Result.Left + OffScreen.Canvas.TextWidth(
    FloatToStr(FMaxY));
end;

{ GetDrawRect - get area for drawing }
function TXYChart.GetDrawRect: TRect;
begin
  { get chart rect }
  Result := GetChartRect;
  { offset inwards 3 and 4 percent }
  Result.Left := Result.Left + ((4 * Width) div 100);
  Result.Top := Result.Top + ((3 * Width) div 100);
  Result.Right := Result.Right - ((4 * Width) div 100);
  Result.Bottom := Result.Bottom - ((3 * Width) div 100);
end;

{ GetBarWidth - get width of bars for bar chart }
function TXYChart.GetBarWidth: integer;
var
  DrawRect: TRect;
begin
  DrawRect := GetDrawRect;
  Result := (((DrawRect.Right-DrawRect.Left) div FDataList.Count) div 2);
end;

{ ScaleChart - calculate extrema }
procedure TXYChart.ScaleChart;
var
  YDiff, XDiff, Mean: Extended;

  function XMaxima: Double;
  var
    i: integer;
  begin
    Result := FDataList.GetDataPoint(0).X;
    for i := 0 to FDataList.Count-1 do
      if FDataList.GetDataPoint(i).X > Result then Result := FDataList.GetDataPoint(i).X;
  end;

  function YMaxima: Double;
  var
    i: integer;
  begin
    Result := FDataList.GetDataPoint(0).Y;
    for i := 0 to FDataList.Count-1 do
      if FDataList.GetDataPoint(i).Y > Result then Result := FDataList.GetDataPoint(i).Y;
  end;

  function XMinima: Double;
  var
    i: integer;
  begin
    Result := FDataList.GetDataPoint(0).X;
    for i := 0 to FDataList.Count-1 do
      if FDataList.GetDataPoint(i).X < Result then Result := FDataList.GetDataPoint(i).X;
  end;

  function YMinima: Double;
  var
    i: integer;
  begin
    Result := FDataList.GetDataPoint(0).Y;
    for i := 0 to FDataList.Count-1 do
      if FDataList.GetDataPoint(i).Y < Result then Result := FDataList.GetDataPoint(i).Y;
  end;

  function TruncFloat(Value: Extended): Extended;
  var
    FloatRec: TFloatRec;
    Temp: string;
  begin
    FloatToDecimal(FloatRec, Value, fvExtended, 12, 9999);
    Temp := FloatToStrF(Value, ffExponent, 1, FloatRec.Exponent);
    Result := StrToFloat(Temp);
  end;

begin
  if FDataList.Count <> 0 then
  begin
    { set extreme values }
    FMinX := Floor(XMinima);
    FMaxX := Ceil(XMaxima);
    FMinY := Floor(YMinima);
    FMaxY := Ceil(YMaxima);
    if FMaxX-FMinX <= 2 then
    begin
      XDiff := TruncFloat(XMaxima-XMinima);
      Mean := TruncFloat(((XMaxima+XMinima)/2));
      FMinX := Mean - XDiff;
      FMaxX := Mean + XDiff;
      if FMaxX = FMinX then
      begin
        FMaxX := FMaxX + 1;
        FMinX := FMinX - 1;
      end;
    end;
    if FMaxY-FMinY <= 2 then
    begin
      YDiff := TruncFloat(YMaxima-YMinima);
      Mean := TruncFloat(((YMaxima+YMinima)/2));
      FMinY := Mean - YDiff;
      FMaxY := Mean + YDiff;
      if FMaxY = FMinY then
      begin
        FMaxY := FMaxY + 1;
        FMinY := FMinY - 1;
      end;
    end;
  end;
  if Assigned(FOnScaling) then FOnScaling(Self, FMinX, FMaxX, FMinY, FMaxY);
end;

{ procedure to draw grids if desired }
procedure TXYChart.DrawGrid(Rect: TRect);
var
  i: integer;
  DrawRect: TRect;
  Xinterval: integer;
  Yinterval: integer;
  GraphWd: integer;
  GraphHt: integer;

  function GetDivisionWidth(Max, Min: integer): integer;
  var
    Length: integer;
  begin
    Length := Max-Min;
    Result := Length div (OffScreen.Canvas.TextWidth(FloatToStr(FMaxX)) * 3);
  end;

  function GetDivisionHeight(Max, Min: integer): integer;
  var
    Length: integer;
  begin
    Length := Max-Min;
    Result := Length div (OffScreen.Canvas.TextHeight('A') * 2);
  end;

  procedure DoRightXDivision(Amount: integer);
  begin
    with OffScreen.Canvas Do
    begin
      MoveTo(DrawRect.Right-Amount, ChartRect.Top+BorderWidth);
      LineTo(DrawRect.Right-Amount, ChartRect.Bottom-BorderWidth);
    end;
  end;

  procedure DoRightYDivision(Amount: integer);
  begin
    with OffScreen.Canvas Do
    begin
      MoveTo(ChartRect.Left+BorderWidth, DrawRect.Top + Amount);
      LineTo(ChartRect.Right-BorderWidth, DrawRect.Top + Amount);
    end;
  end;

begin
  if FGrid = gtNone then Exit;
  DrawRect := GetDrawRect;
  GraphWd := (DrawRect.Right - DrawRect.Left);
  GraphHt := (DrawRect.Bottom - DrawRect.Top);
  Xinterval := GetDivisionWidth(DrawRect.Right, DrawRect.Left);
  Yinterval := GetDivisionHeight(DrawRect.Bottom, DrawRect.Top);
  with OffScreen.Canvas do
  begin
    if FChartColor <> clSilver then Pen.Color := clSilver
      else Pen.Color := clGray;
    case FGrid of
      gtBoth, gtVert:
        case XInterval of
        0,1,2: begin
            DoRightXDivision(GraphWd);
            DoRightXDivision(GraphWd Div 2);
            DoRightXDivision(0);
          end;
        3: begin
            DoRightXDivision(0);
            DoRightXDivision(GraphWd Div 4);
            DoRightXDivision(GraphWd Div 2);
            DoRightXDivision(GraphWd Div 4*3);
            DoRightXDivision(GraphWd);
          end;
        4: begin
          for I := 0 to 10 do
            if not Odd(I) then
              DoRightXDivision(Round(GraphWd * I/10));
          end;
        else
          for I := 0 To 10 do
            DoRightXDivision(Round(GraphWd * I/10));
        end;
    end;
    case FGrid of
      gtBoth, gtHorz:
        case YInterval of
        0,1,2: begin
            DoRightYDivision(GraphHt);
            DoRightYDivision(GraphHt Div 2);
            DoRightYDivision(0);
          end;
        3: begin
            DoRightYDivision(0);
            DoRightYDivision(GraphHt Div 4);
            DoRightYDivision(GraphHt Div 2);
            DoRightYDivision(GraphHt Div 4*3);
            DoRightYDivision(GraphHt);
          end;
        4: begin
          for I := 0 to 10 do
            if not Odd(I) then
              DoRightYDivision(Round(GraphHt * I/10));
          end;
        else
          for I := 0 To 10 do
            DoRightYDivision(Round(GraphHt * I/10));
        end;
    end;
  end;
end;

procedure TXYChart.DrawXScale;
var
  i: integer;
  DrawRect: TRect;
  interval: integer;
  GraphHt: integer;
  Range: Double;

  function GetDivisionWidth(Max, Min: integer): integer;
  var
    Length: integer;
  begin
    Length := Max-Min;
    Result := Length div (OffScreen.Canvas.TextWidth(FloatToStr(FMaxX)) * 3);
  end;

  procedure DoRightDivision(Amount: integer; Value: Double);
  var
    S: string;
  begin
    with OffScreen.Canvas Do
    begin
      MoveTo(DrawRect.Right-Amount, ChartRect.Bottom-BorderWidth-5);
      LineTo(DrawRect.Right-Amount, ChartRect.Bottom-BorderWidth);
      S := FloatToStr(Value);
      if Assigned(FOnDrawScales) then FOnDrawScales(Self, atXAxis, S);
      TextOut(DrawRect.Right-(TextWidth(S) div 2)-Amount, ChartRect.Bottom+3, S);
    end;
  end;

begin
  DrawRect := GetDrawRect;
  GraphHt := (DrawRect.Right - DrawRect.Left);
  Range := FMaxX-FMinX;
  interval := GetDivisionWidth(DrawRect.Right, DrawRect.Left);
  with OffScreen.Canvas do
  begin
    Pen.Color := clBlack;
    Brush.Color := FBackGround;
    case interval of
     0: begin
          MoveTo(DrawRect.Left+(GraphHt div 2), ChartRect.Bottom-BorderWidth-5);
          LineTo(DrawRect.Left+(GraphHt div 2), ChartRect.Bottom-BorderWidth);
        end;
     1: DoRightDivision(GraphHt Div 2, Range / 2 + FMinX);
     2: begin
          DoRightDivision(GraphHt, FMinX);
          DoRightDivision(GraphHt Div 2, Range / 2+FMinX);
          DoRightDivision(0, FMaxX);
        end;
     3: begin
          DoRightDivision(0, FMaxX);
          DoRightDivision(GraphHt Div 4, Range / 4*3+FMinX);
          DoRightDivision(GraphHt Div 2, Range / 2+FMinX);
          DoRightDivision(GraphHt Div 4*3, Range / 4+FMinX);
          DoRightDivision(GraphHt, FMinX);
        end;
     4: begin
          for I := 0 to 10 do
            if not Odd(I) then
              DoRightDivision(Round(GraphHt * I/10), Range*(10-I)/10+FMinX);
        end;
     else
        for I := 0 To 10 do
          DoRightDivision(Round(GraphHt* I/10), Range*(10-I)/10+FMinX);
    end; {Case}
  end;
end;

procedure TXYChart.DrawYScale;
var
  i: integer;
  DrawRect: TRect;
  interval: LongInt;
  GraphHt: integer;
  Range: Double;

  function GetDivisionHeight(Max, Min: integer): integer;
  var
    Length: integer;
  begin
    Length := Max-Min;
    Result := Length div (OffScreen.Canvas.TextHeight('A') * 2);
  end;

  procedure DoRightDivision(Amount: integer; Value: Double);
  var
    S: string;
  begin
    with OffScreen.Canvas Do
    begin
      MoveTo(ChartRect.Left, DrawRect.Top+Amount);
      LineTo(ChartRect.Left+5, DrawRect.Top+Amount);
      S := FloatToStr(Value);
      if Assigned(FOnDrawScales) then FOnDrawScales(Self, atYAxis, S);
      TextOut(ChartRect.Left-3-TextWidth(S), DrawRect.Top+Amount-(TextHeight(S) div 2), S);
    end;
  end;

begin
  DrawRect := GetDrawRect;
  GraphHt := (DrawRect.Bottom - DrawRect.Top);
  Range := FMaxY-FMinY;
  interval := GetDivisionHeight(DrawRect.Bottom, DrawRect.Top);
  with OffScreen.Canvas do
  begin
    Pen.Color := clBlack;
    Brush.Color := FBackGround;
    case interval of
     0: begin
          MoveTo(ChartRect.Left, DrawRect.Top + (GraphHt div 2));
          LineTo(ChartRect.Left+5, DrawRect.Top + (GraphHt div 2));
        end;
     1: DoRightDivision(GraphHt div 2,Range / 2+FMinY);
     2: begin
          DoRightDivision(GraphHt, FMinY);
          DoRightDivision(GraphHt div 2, Range / 2+FMinY);
          DoRightDivision(0, FMaxY);
        end;
     3: begin
          DoRightDivision(GraphHt, FMinY);
          DoRightDivision(GraphHt div 4,Range / 4*3+FMinY);
          DoRightDivision(GraphHt div 2,Range / 2+FMinY);
          DoRightDivision(GraphHt div 4*3,Range / 4+FMinY);
          DoRightDivision(0, FMaxY);
        end;
     4: begin
          for I := 0 to 10 do
            if not Odd(I) then
              DoRightDivision(Round(GraphHt * I/10), Range*(10-I)/10+FMinY);
        end;
     else
        for I := 0 To 10 do
          DoRightDivision(Round(GraphHt* I/10), Range*(10-I)/10+FMinY);
    end; {Case}
  end;
end;

{ NormalizePtX - transform  X-axis data to chart dimensions }
function TXYChart.NormalizePtX(APoint: Double): Double;
var
  DrawRect: TRect;
  W: integer;
begin
  Result := APoint;
  { get width of drawing area }
  DrawRect := GetDrawRect;
  W := DrawRect.Right-DrawRect.Left;
  { if axis range <> 0 - otherwise div by zero }
  if FMaxX-FMinX <> 0 then
    { make data point a relative fraction of
      the chart draw rect }
    Result := (APoint-FMinX)*W/(FMaxX-FMinX);
end;

{ NormalizePtY - transform  Y-axis data to chart dimensions }
function TXYChart.NormalizePtY(APoint: Double): Double;
var
  DrawRect: TRect;
  H: integer;
begin
  Result := APoint;
  { get height of drawing area }
  DrawRect := GetDrawRect;
  H := DrawRect.Bottom-DrawRect.Top;
  { if axis range <> 0 - otherwise div by zero }
  if FMaxX-FMinX <> 0 then
    { make data point a relative fraction of
      the chart draw rect }
    Result := (APoint-FMinY)*H/(FMaxY-FMinY);
end;

procedure TXYChart.DrawDataPoints;
var
  i: integer;
  DrawRect: TRect;
  X, Y: Double;
  Finished: boolean;
  AColor: TColor;
begin
  ScaleChart;
  DrawRect := GetDrawRect;
  with OffScreen.Canvas do
  begin
    Pen.Color := clBlack;
    { move pen to 1st data point }
    if FDataList.Count <> 0 then
    begin
      MoveTo(DrawRect.Left+Round(NormalizePtX(FDataList.GetDataPoint(0).X)),
        DrawRect.Bottom-Round(NormalizePtY(FDataList.GetDataPoint(0).Y)));
      { iterate through data }
      for i := 0 to FDataList.Count-1 do
      begin
        case FChartType of
          ctLine: LineTo(DrawRect.Left+Round(NormalizePtX(FDataList.GetDataPoint(i).X)),
            DrawRect.Bottom-Round(NormalizePtY(FDataList.GetDataPoint(i).Y)));
          ctBar: begin
            Brush.Style := bsSolid;
            Brush.Color := clNavy;
            Rectangle(DrawRect.Left+Ceil(NormalizePtX(FDataList.GetDataPoint(i).X))-BarWidth,
              DrawRect.Bottom-Round(NormalizePtY(FDataList.GetDataPoint(i).Y)),
                DrawRect.Left+Floor(NormalizePtX(FDataList.GetDataPoint(i).X))+BarWidth,
                  DrawRect.Bottom);
          end;
          ctScatter: Ellipse(DrawRect.Left+Round(NormalizePtX(FDataList.GetDataPoint(i).X)-3),
            DrawRect.Bottom-Round(NormalizePtY(FDataList.GetDataPoint(i).Y)-3),
              DrawRect.Left+Round(NormalizePtX(FDataList.GetDataPoint(i).X)+3),
                DrawRect.Bottom-Round(NormalizePtY(FDataList.GetDataPoint(i).Y)+3));
          ctXY: begin
            LineTo(DrawRect.Left+Round(NormalizePtX(FDataList.GetDataPoint(i).X)),
              DrawRect.Bottom-Round(NormalizePtY(FDataList.GetDataPoint(i).Y)));
            Ellipse(DrawRect.Left+Round(NormalizePtX(FDataList.GetDataPoint(i).X)-3),
              DrawRect.Bottom-Round(NormalizePtY(FDataList.GetDataPoint(i).Y)-3),
                DrawRect.Left+Round(NormalizePtX(FDataList.GetDataPoint(i).X)+3),
                  DrawRect.Bottom-Round(NormalizePtY(FDataList.GetDataPoint(i).Y)+3));
          end;
        end;
      end;
      { Set finished to true }
      Finished := true;
      { set tag - counter - to 0 }
      Self.Tag := 0;
      { set X and Y to values below Minima }
      X := FMinX-10;
      Y := FMinY-10;
      repeat
        if Assigned(FOnNeedLinesX) then
        begin
          { trigger event }
          FOnNeedLinesX(Self, X, AColor, Finished);
          X := NormalizePtX(X);
          if (DrawRect.Left+Round(X) < DrawRect.Right) and
            (DrawRect.Left+Round(X) > DrawRect.Left) then with OffScreen.Canvas do
          begin
            Pen.Color := AColor;
            { draw line }
            MoveTo(DrawRect.Left+Round(X), DrawRect.Top);
            LineTo(DrawRect.Left+Round(X), DrawRect.Bottom);
          end;
          { increment tag - counter }
          Self.Tag := Self.Tag + 1;
        end;
      until Finished = true;
      { reset counter - finished is already true }
      Self.Tag := 0;
      repeat
        if Assigned(FOnNeedLinesY) then
        begin
          FOnNeedLinesY(Self, Y, AColor, Finished);
          Y := NormalizePtY(Y);
          if (DrawRect.Bottom-Round(Y) > DrawRect.Top) and
            (DrawRect.Bottom-Round(Y) < DrawRect.Bottom) then with OffScreen.Canvas do
          begin
            Pen.Color := AColor;
            { draw line }
            MoveTo(DrawRect.Left, DrawRect.Bottom-Round(Y));
            LineTo(DrawRect.Right, DrawRect.Bottom-Round(Y));
          end;
          { increment tag - counter }
          Self.Tag := Self.Tag + 1;
        end;
      until Finished = true;
      { reset pen color }
      Pen.Color := clBlack;
    end;
  end;
end;

{ register component on HomeGrown page }
procedure Register;
begin
  RegisterComponents('HomeGrown', [TXYChart]);
end;

end.